# Copyright 2020 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Destroy action module."""

import logging

from mrack.context import global_context
from mrack.errors import MetadataError
from mrack.transformers import transformers

logger = logging.getLogger(__name__)


class Action:
    """Base Action."""

    def __init__(self, config=None, metadata=None, db_driver=None):
        """Initialize the action."""
        self._config = config or global_context.PROV_CONFIG
        self._metadata = metadata or global_context.METADATA
        self._db_driver = db_driver or global_context.DB
        self._transformers = {}

    async def _get_transformer(self, provider_name):
        """Get a transformer by name, initialize a new one if not yet done."""
        transformer = self._transformers.get(provider_name)
        if not transformer:
            transformer = transformers.get(provider_name)
            await transformer.init(self._config, self._metadata)
            if not transformer:
                raise MetadataError(f"Invalid provider: {provider_name}")
            self._transformers[provider_name] = transformer
        return transformer


class DBAction:
    """Base Action."""

    def __init__(
        self,
        db_driver=None,
    ):
        """Initialize the database action."""
        self._db_driver = db_driver or global_context.DB



# Copyright 2020 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Podman provider module."""

import asyncio
import logging
import os
from datetime import datetime, timedelta

from mrack.errors import ProvisioningError, ServerNotFoundError
from mrack.host import STATUS_ACTIVE, STATUS_DELETED, STATUS_ERROR, STATUS_OTHER
from mrack.providers.provider import STRATEGY_ABORT, Provider
from mrack.providers.utils.podman import Podman
from mrack.utils import object2json

logger = logging.getLogger(__name__)

PROVISIONER_KEY = "podman"


class PodmanProvider(Provider):
    """Podman provisioning provider."""

    def __init__(self):
        """Initialize provider."""
        super().__init__()
        self._name = PROVISIONER_KEY
        self.dsp_name = "Podman"
        self.max_retry = 1  # for retry strategy
        self.podman = Podman()
        self.status_map = {
            STATUS_ACTIVE: STATUS_ACTIVE,
            STATUS_DELETED: STATUS_DELETED,
            STATUS_ERROR: STATUS_ERROR,
        }

    async def init(
        self,
        container_images,
        default_network,
        ssh_key,
        container_options,
        extra_commands,
        strategy=STRATEGY_ABORT,
        max_retry=1,
    ):
        """Initialize Podman provider with data from config."""
        logger.info(f"{self.dsp_name} Initializing provider")
        login_start = datetime.now()
        self.strategy = strategy
        self.max_retry = max_retry
        self.images = container_images
        self.default_network = default_network
        self.ssh_key = ssh_key
        self.podman_options = container_options
        self.extra_commands = extra_commands
        login_end = datetime.now()
        login_duration = login_end - login_start
        logger.info(f"{self.dsp_name} Init duration {login_duration}")

    async def prepare_provisioning(self, reqs):
        """Prepare podman network and pull missing images to prepare provisioning."""
        image_list = set()
        network_list = set()

        for req in reqs:
            # First create list of podman network(s) to be created
            # podman network(s) should be always created no matter what
            network = req.get(
                "network", f"{self.default_network}-{req['domain'].replace('.', '-')}"
            )
            # store the used network per requirement to be later accessed
            req["network"] = network
            network_list.add(network)

            # Prepare image_list to be pulled later
            image_list.add(req["image"])

        if network_list:
            logger.info(f"{self.dsp_name} Preparing network(s) {network_list}")
            awaitables = []
            for network in network_list:
                if not await self.podman.network_exists(network):
                    awaitables.append(self.podman.network_create(network))

            network_results = await asyncio.gather(*awaitables)
            success = all(network_results)

            if not success:
                logger.error(f"{self.dsp_name} Creation of networks failed")
                return False

        logger.info(f"{self.dsp_name} Pulling missing images {image_list}")

        awaitables = []
        for image in image_list:
            if not await self.podman.image_exists(image):
                logger.debug(f"{self.dsp_name} Pull of image '{image}' required")
                awaitables.append(self.podman.pull(image))

        pull_results = await asyncio.gather(*awaitables)
        success = all(pull_results)

        if success:
            logger.info(f"{self.dsp_name} All required images present")
        else:
            raise ProvisioningError("Pulling of missing images failed", self.dsp_name)

        return success

    async def validate_hosts(self, reqs):
        """Validate that host requirements are well specified."""
        return bool(reqs)  # TODO

    async def can_provision(self, hosts):
        """Check that provider has enough resources to provision hosts."""
        return bool(hosts)  # TODO

    async def utilization(self):
        """Check percentage utilization of given provider."""
        return 0

    async def create_server(self, req):
        """Request and create resource on selected provider."""
        hostname = req["name"]
        log_msg_start = f"{self.dsp_name} [{hostname}]"
        logger.info(f"{log_msg_start} Creating container for host")

        image = req["image"]
        network = req.get("network")  # preparation method should set this value
        if not network:
            logger.error(
                f"{log_msg_start} Failed to load network requirement from: {req}"
            )
            raise ProvisioningError(
                "Could not set up podman network for some host(s)", req
            )

        try:
            container_id = await self.podman.run(
                image,
                hostname,
                network,
                extra_options=self.podman_options,
                remove_at_stop=True,
            )
        except ProvisioningError as p_error:
            raise ProvisioningError(p_error, req) from p_error

        return (container_id, req)

    async def wait_till_provisioned(self, resource):
        """Wait till resource is provisioned."""
        # access fist item from tuple resource which should be id
        cont_id, req = resource
        log_msg_start = f"{self.dsp_name} [{req.get('name')}]"

        start = datetime.now()
        timeout = 20
        timeout_time = start + timedelta(minutes=timeout)

        while datetime.now() < timeout_time:
            try:
                server = await self.podman.inspect(cont_id)[0]
            except ProvisioningError as err:
                logger.error(f"{log_msg_start} {object2json(err)}")
                raise ServerNotFoundError(cont_id) from err

            if server["State"]["Running"] or server["State"]["Error"]:
                break
            await asyncio.sleep(1)

        done_time = datetime.now()
        prov_duration = (done_time - start).total_seconds()

        if datetime.now() >= timeout_time:
            logger.warning(
                f"{log_msg_start} Container {cont_id} was not provisioned "
                f"within a timeout of {timeout} mins"
            )
        else:
            logger.info(
                f"{log_msg_start} Container {cont_id} "
                f"was provisioned in {prov_duration:.1f}s"
            )

        logger.debug(f"{log_msg_start} Resource: {object2json(server)}")

        with open(os.path.expanduser(self.ssh_key), "r", encoding="utf-8") as key_file:
            key_content = key_file.read()

        if not await self.podman.exec_command(cont_id, "mkdir -p /root/.ssh/"):
            raise ProvisioningError(
                f"Could not copy public key to container {cont_id}", self.dsp_name
            )

        if not await self.podman.exec_command(
            cont_id, f'echo "{key_content}" >> /root/.ssh/authorized_keys'
        ):
            raise ProvisioningError(
                f"Could not copy public key to container {cont_id}", self.dsp_name
            )

        for command in self.extra_commands:
            if not await self.podman.exec_command(cont_id, command):
                raise ProvisioningError(
                    f"Failed to run '{command}' in container {cont_id}", self.dsp_name
                )

        server.update({"mrack_req": req})

        return server, req

    async def delete_host(self, host_id, host_name):
        """Delete provisioned host."""
        # if there is no container we do nothing
        log_msg_start = f"{self.dsp_name} [{host_name}]"
        if not host_id:
            logger.debug(f"{log_msg_start} Container is not created yet, skipping.")
            return False

        # first we inspect the container to find its networks
        insp_data = await self.podman.inspect(host_id)
        networks = insp_data[0]["NetworkSettings"]["Networks"] if insp_data else []
        # then we destroy the container
        logger.info(f"{log_msg_start} Removing container {host_id}")
        deleted = await self.podman.rm(host_id, force=True)
        # after that we cleanup the podman network
        for net in networks:
            if await self.podman.network_remove(net):
                logger.info(f"{log_msg_start} Removed network '{net}'")

        return deleted

    def get_status(self, state):
        """Read status from inspect State object."""
        if state.get("Running"):
            status = STATUS_ACTIVE
        elif state.get("Dead"):
            status = STATUS_DELETED
        elif state.get("Error"):
            status = STATUS_ERROR
        else:
            status = STATUS_OTHER

        return status

    def prov_result_to_host_data(self, prov_result, req):
        """Get needed host information from podman provisioning result."""
        log_msg_start = f"{self.dsp_name} [{req.get('name')}]"

        result = {}
        result["id"] = prov_result.get("Id")
        result["name"] = prov_result["Config"]["Hostname"]

        network_set = prov_result.get("NetworkSettings")

        result["addresses"] = []
        if network_set:
            try:  # TODO self.network_name
                for net in network_set["Networks"]:
                    result["addresses"].append(
                        network_set["Networks"][net]["IPAddress"]
                    )
            except KeyError as kerror:
                raise ProvisioningError(
                    f"{log_msg_start} Container state improper"
                ) from kerror

        status = self.get_status(prov_result.get("State"))
        error_obj = None
        if status == STATUS_ERROR:
            error_obj = prov_result.get("State")

        result["fault"] = error_obj
        result["status"] = status
        result["os"] = prov_result.get("mrack_req").get("os")
        result["group"] = prov_result.get("mrack_req").get("group")

        return result

    def to_host(self, provisioning_result, req, username=None):
        """Transform provisioning result into Host object."""
        # for containers use always root user
        return super().to_host(provisioning_result, req, username="root")



# Copyright 2020 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""List action module."""
import logging

from mrack.actions.action import DBAction

logger = logging.getLogger(__name__)


class List(DBAction):
    """List action.

    List hosts from DB.
    """

    def list(self):
        """Execute the List action."""
        hosts = self._db_driver.hosts.values()

        for host in hosts:
            logger.info(host)



# Copyright 2020 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions for output modules."""

import socket
from socket import error as socket_error

from mrack.utils import find_value_in_config_hierarchy


def resolve_hostname(ip_addr):
    """Resolve IP address to hostname."""
    try:
        return socket.gethostbyaddr(ip_addr)[0]
    except socket_error:
        return None


def get_external_id(host, meta_host, config):
    """
    Get host's external ID.

    That can be its resolvable DNS name (from IP) or the IP - based on provider or
    host configuration (key: resolve_host, default True).

    IP is used as fallback if the desired is not available.
    """
    resolve_ip = find_value_in_config_hierarchy(
        config, host.provider.name, host, meta_host, "resolve_host", None, None, True
    )

    external_id = host.ip_addr
    if resolve_ip:
        external_id = resolve_hostname(host.ip_addr) or host.ip_addr
    return external_id



# Copyright 2020 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Static transformer module."""
import typing
from copy import deepcopy

from mrack.transformers.transformer import Transformer

CONFIG_KEY = "static"


class StaticTransformer(Transformer):
    """
    Static transformer.

    Does almost no operation as there is nothing to provision.
    """

    _config_key = CONFIG_KEY
    _required_config_attrs: typing.List[str] = []
    _required_host_attrs = ["name", "os", "group", "ip"]

    def create_host_requirement(self, host):
        """Create single input for Static provisioner."""
        self.dsp_name = "Static"
        return deepcopy(host)



# Copyright 2021 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Global mrack context object."""
import os
from typing import Dict

from mrack.config import MrackConfig, ProvisioningConfig
from mrack.dbdrivers.file import FileDBDriver
from mrack.errors import ConfigError
from mrack.utils import NoSuchFileHandler, load_yaml


class GlobalContext:
    """Global context class to store the mrack configuration."""

    def __init__(self) -> None:
        """Init empty GlobalContext class."""
        self.database = None
        self.mrack_conf = None
        self.provisioning_config = None
        self.metadata: Dict = {}

    @property
    def DB(self):  # pylint: disable=invalid-name
        """Get FileDBDriver object."""
        return self.database

    @property
    def CONFIG(self):  # pylint: disable=invalid-name
        """Get MrackConfig object."""
        return self.mrack_conf

    @property
    def PROV_CONFIG(self):  # pylint: disable=invalid-name
        """Get ProvisioningConfig object."""
        return self.provisioning_config

    @property
    def METADATA(self):  # pylint: disable=invalid-name
        """Get ProvisioningConfig object."""
        return self.metadata

    def init(self, mrack_config, provisioning_config=None, db_file=None):
        """Initialize Global Context object with all needed values."""
        self._init_mrack_config(mrack_config)
        self.mrack_conf.load()

        db_path = db_file or self.mrack_conf.db_path(default="./.mrackdb.json")
        p_config_path = provisioning_config or self.mrack_conf.provisioning_config_path(
            default="./provisioning-config.yaml"
        )

        self._init_db(db_path)
        self._init_prov_config(p_config_path)

    def _init_db(self, path):
        """Initialize file database."""
        self.database = FileDBDriver(path)

    @NoSuchFileHandler(error="Provisioning config file not found: {path}")
    def _init_prov_config(self, path):
        """Load and initialize provisioning configuration."""
        self.provisioning_config = ProvisioningConfig(load_yaml(path))

    def init_metadata(self, user_defined_path):
        """Load and initialize job metadata."""
        meta_path = user_defined_path or self.mrack_conf.metadata_path()

        if not meta_path:
            raise ConfigError("Job metadata file path not provided.")
        if not os.path.exists(meta_path):
            raise ConfigError(f"Job metadata file not found: {meta_path}")

        self.metadata = load_yaml(meta_path)

    def _init_mrack_config(self, mrack_config_path):
        """Load and initialize mrack configuration."""
        self.mrack_conf = MrackConfig(mrack_config_path)


# Singleton context holder
global_context = GlobalContext()



# Copyright 2020 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Additional client API wrappers for OpenStack."""

from asyncopenstackclient import NovaClient
from asyncopenstackclient.client import Client


class ExtraNovaClient(NovaClient):
    """Extension of NovaClient to provide additional methods.

    Added methods:
    * limits.show
    * quota.show
    * usage.show
    """

    def __init__(self, session=None, api_url=None):
        """Add new objects."""
        super().__init__(session, api_url)
        self.resources.extend(["limits", "quota", "usage"])

    async def init_api(self, timeout=60):
        """Initialize API.

        to add additional methods
        """
        await super().init_api(timeout)
        self.api.limits.actions["show"] = {"method": "GET", "url": "limits"}
        self.api.quota.actions["show"] = {"method": "GET", "url": "os-quota-sets/{}"}
        self.api.usage.actions["show"] = {
            "method": "GET",
            "url": "os-simple-tenant-usage/{}",
        }
        self.api.limits.add_action("show")
        self.api.quota.add_action("show")
        self.api.usage.add_action("show")


class NeutronClient(Client):
    """Client API for working with Neutron (Networks).

    Available methods:
    * network.list - get all networks
    * ip.list - get network availibilities
    """

    def __init__(self, session=None, api_url=None):
        """Add new objects."""
        super().__init__("neutron", ["network", "ip"], session, api_url)

    async def init_api(self, timeout=60):
        """Initialize API.

        to add additional methods
        """
        await super().init_api(timeout)
        self.api.network.actions["list"] = {"method": "GET", "url": "networks"}
        self.api.ip.actions["list"] = {
            "method": "GET",
            "url": "network-ip-availabilities",
        }
        self.api.network.add_action("list")
        self.api.ip.add_action("list")



# Copyright 2020 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for working with podman."""

import json
import logging
import subprocess

from mrack.errors import ProvisioningError
from mrack.utils import exec_async_subprocess

logger = logging.getLogger(__name__)


class Podman:
    """Async wrapper supporting most basic podman calls."""

    def __init__(self, program="podman"):
        """Init the instance."""
        self.program = program
        self.dsp_name = program.capitalize()

    async def _run_podman(self, args, raise_on_err=True):
        """Util method to execute podman process."""
        try:
            return await exec_async_subprocess(self.program, args, raise_on_err)
        except ProvisioningError as p_error:
            try:
                err_str = str(p_error).split("Error:")[1].strip()
            except IndexError:
                err_str = str(p_error)

            raise ProvisioningError(err_str) from p_error

    async def run(
        self,
        image,
        hostname=None,
        network=None,
        extra_options=None,
        remove_at_stop=False,
    ):
        """Run a container."""
        args = ["run", "-dti"]

        for opt in extra_options:
            if isinstance(extra_options[opt], list):
                for item in extra_options[opt]:
                    args.extend([opt, item])
            else:
                args.extend([opt, extra_options[opt]])

        if remove_at_stop:
            args.append("--rm")

        if network:
            args.extend(["--network", network])

        if hostname:
            args.extend(["-h", hostname])
            args.extend(["--name", f"{hostname.replace('.', '-')}-{network}"])

        args.append(image)
        stdout, _stderr, _process = await self._run_podman(args)
        container_id = stdout.strip()
        return container_id

    async def inspect(self, container_id):
        """Inspects a container returns data loaded from JSON structure."""
        args = ["inspect", container_id]
        stdout, _stderr, _process = await self._run_podman(args, raise_on_err=False)
        inspect_data = json.loads(stdout)
        return inspect_data

    async def rm(self, container_id, force=False):  # pylint: disable=invalid-name
        """Remove a container."""
        args = ["rm"]
        if force:
            args.append("-f")
        args.append(container_id)
        _stdout, stderr, process = await self._run_podman(args, raise_on_err=False)

        if stderr:
            logger.debug(f"{self.dsp_name} {stderr.strip()}")

        return process.returncode == 0

    async def stop(self, container_id, time=0):
        """Remove a container."""
        args = ["stop"]
        if time:
            args.append("--time")
            args.append(time)

        args.append(container_id)
        _stdout, _stderr, process = await self._run_podman(args, raise_on_err=False)
        return process.returncode == 0

    async def exec_command(self, container_id, command):
        """Execute command in selected container."""
        args = ["exec", container_id, "sh", "-c"]
        args.append(command)
        _stdout, _stderr, process = await self._run_podman(args, raise_on_err=False)
        return process.returncode == 0

    async def network_exists(self, network):
        """Check the existence of podman network on system using inspect command."""
        args = ["network", "inspect", network]
        _stdout, _stderr, inspect = await self._run_podman(args, raise_on_err=False)
        return inspect.returncode == 0

    async def network_create(self, network):
        """Create a podman network if it does not exist."""
        if await self.network_exists(network):
            logger.debug(f"{self.dsp_name} Network '{network}' is present")
            return 0

        logger.info(f"{self.dsp_name} Creating podman network '{network}'")
        args = ["network", "create", network]
        _stdout, _stderr, process = await self._run_podman(args, raise_on_err=False)

        return process.returncode == 0

    async def network_remove(self, network):
        """Remove a podman network if it does exist."""
        if not await self.network_exists(network):
            logger.debug(f"{self.dsp_name} Network '{network}' does not exists")
            return True

        args = ["network", "remove", network]
        _stdout, _stderr, process = await self._run_podman(args, raise_on_err=False)

        return process.returncode == 0

    async def pull(self, image):
        """Pull a container image."""
        args = ["pull", image]
        logger.info(
            f"{self.dsp_name} Pulling image '{image}'. This may take a while..."
        )
        _stdout, _stderr, process = await self._run_podman(args, raise_on_err=False)
        if process.returncode == 0:
            logger.info(f"{self.dsp_name} Pull of image '{image}' succeeded")
        else:
            logger.error(f"{self.dsp_name} Pull of image '{image}' failed")

        return process.returncode == 0

    async def image_exists(self, image):
        """Check if a container image exists in local storage."""
        args = ["image", "exists", image]
        _stdout, _stderr, process = await self._run_podman(args, raise_on_err=False)
        return process.returncode == 0

    def interactive(self, container_id):
        """Create interactive session."""
        args = [self.program, "exec", "-ti", container_id, "bash"]
        try:
            subprocess.run(args, text=True, check=True)
        except subprocess.CalledProcessError as callerr:
            if callerr.returncode != 130:
                raise  # when it was not killed by ctrl + D (signal 2)



# Copyright 2020 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""AWS Provider interface."""

import asyncio
import logging
import secrets
from copy import deepcopy
from datetime import datetime
from random import shuffle

import boto3
from botocore.exceptions import ClientError, NoCredentialsError, NoRegionError
from dateutil import parser

from mrack.errors import NotAuthenticatedError, ProvisioningError, ValidationError
from mrack.host import STATUS_ACTIVE, STATUS_DELETED, STATUS_ERROR, STATUS_PROVISIONING
from mrack.providers.provider import STRATEGY_ABORT, Provider
from mrack.utils import object2json

logger = logging.getLogger(__name__)

PROVISIONER_KEY = "aws"


class AWSProvider(Provider):
    """AWS Provider."""

    def __init__(self):
        """Object initialization."""
        super().__init__()
        self._name = PROVISIONER_KEY
        self.dsp_name = "AWS"
        self.ssh_key = None
        self.instance_tags = None
        self.max_retry = 1  # for retry strategy
        self.subnets_capacity = {}
        self.status_map = {
            "running": STATUS_ACTIVE,
            "pending": STATUS_PROVISIONING,
            "terminated": STATUS_DELETED,
            "error": STATUS_ERROR,
            # there is much more we can treat it as STATUS_OTHER, see statuses:
            # pending | running | shutting-down | terminated | stopping | stopped
        }

    @property
    def name(self):
        """Get provider name."""
        return self._name

    async def init(
        self,
        ssh_key,
        instance_tags,
        strategy=STRATEGY_ABORT,
        max_retry=1,
    ):
        """Initialize provider with data from AWS."""
        # AWS_CONFIG_FILE=`readlink -f ./aws.key`
        log_msg_start = self.dsp_name
        logger.info(f"{log_msg_start} Initializing provider")
        login_start = datetime.now()
        self.strategy = strategy
        self.max_retry = max_retry
        try:
            self.ec2 = boto3.resource("ec2")
            self.client = boto3.client("ec2")
        except (NoRegionError, NoCredentialsError) as c_err:
            logger.debug(
                f"{log_msg_start} Failed loading credentials file with: {str(c_err)}"
            )
            raise NotAuthenticatedError(
                f"{log_msg_start} failed loading credentials. Load AWS credentials"
                " and try again. E.g.: $ export AWS_CONFIG_FILE=~/aws.key"
            ) from c_err

        self.amis = []
        self.ssh_key = ssh_key
        self.instance_tags = instance_tags
        login_end = datetime.now()
        login_duration = login_end - login_start
        logger.info(f"{log_msg_start} Login duration {login_duration}")

    def raise_image_def_error(self, definition):
        """Raise error that image definition is incorrect."""
        json_str = object2json(definition)
        error = f"Not a valid image definition:\n{json_str}"
        raise ValidationError(error)

    def validate_tags_image_def(self, image_def):
        """Validate that tag definition for image is correct."""
        if not isinstance(image_def, dict) or "tag" not in image_def:
            self.raise_image_def_error(image_def)

        tag_def = image_def.get("tag")
        if (
            not isinstance(tag_def, dict)
            or "name" not in tag_def
            or "value" not in tag_def
            or not isinstance(tag_def.get("name"), str)
            or not isinstance(tag_def.get("value"), str)
        ):
            self.raise_image_def_error(image_def)

        return True

    def get_image(self, req):
        """
        Get a loaded image.

        Does also basic image definition validation.

        Return None if image is not yet loaded.
        """
        log_msg_start = f"{self.dsp_name} [{req.get('name')}]"
        image_def = req.get("image")

        if not image_def:
            raise ValidationError(f"{log_msg_start} Host doesn't have image defined.")

        # by tag
        if isinstance(image_def, dict) and "tag" in image_def:
            self.validate_tags_image_def(image_def)
            tag_def = image_def.get("tag")
            for ami in self.amis:
                if not ami.tags:
                    continue
                for tag in ami.tags:
                    if (
                        tag["Key"] == tag_def["name"]
                        and tag["Value"] == tag_def["value"]
                    ):
                        return ami
        # by AMI ID
        elif isinstance(image_def, str):
            for ami in self.amis:
                if ami.image_id == image_def:
                    return ami
        else:
            raise ValidationError(
                f"{log_msg_start} Invalid image "
                f"definition. Must be 'tags' definition or AMI ID"
            )
        return None

    def load_image(self, req):
        """
        Load AMI information from EC2 based on image requirement.

        If more images match the search then the newest is returned.

        Raises validation error if no image is found.
        """
        log_msg_start = f"{self.dsp_name} [{req.get('name')}]"
        image_def = req.get("image")
        filters = []

        # by tag
        if isinstance(image_def, dict) and "tag" in image_def:
            tag_def = image_def.get("tag")
            name = tag_def["name"]
            filters.append({"Name": f"tag:{name}", "Values": [tag_def["value"]]})

        # by AMI ID
        elif isinstance(image_def, str):
            filters.append({"Name": "image-id", "Values": [image_def]})

        amis = list(self.ec2.images.filter(Filters=filters))

        if not amis:
            raise ValidationError(f"{log_msg_start} Cannot find image for host")

        amis.sort(key=lambda ami: parser.parse(ami.creation_date), reverse=True)
        self.amis.append(amis[0])
        return amis[0]

    def load_images(self, reqs):
        """
        Load AMI images for all requirements.

        Done sequentially, already loaded images are not loaded again. Basically also
        validates that images are available and that their definition is correct.
        """
        for req in reqs:
            ami = self.get_image(req)
            if not ami:
                self.load_image(req)

    async def prepare_provisioning(self, reqs):
        """Prepare provisioning."""
        try:
            self.load_images(reqs)
        except ValidationError as val_err:
            logger.error(val_err)
            return False

        return bool(reqs)

    async def validate_hosts(self, reqs):
        """Validate that host requirements are well specified."""
        return

    async def get_subnet_available_ips(self, subnet_id, log_msg_start):
        """Get number of IPs available in a subnet."""
        try:
            subnet = self.ec2.Subnet(subnet_id)
        except ClientError:
            logger.warning(
                f"{log_msg_start} Error retrieving info from subnet: {subnet_id}"
            )
            return 0

        logger.debug(f"{log_msg_start} Subnet {subnet_id}")
        logger.debug(
            f"{log_msg_start}   available: {subnet.available_ip_address_count}"
        )

        return subnet.available_ip_address_count

    async def can_provision(self, hosts):  # pylint: disable=too-many-branches
        """Check that all host can be provisioned.

        Checks:
        * Available IPv4 addresses are enough
        """
        log_msg_start = self.dsp_name

        ip_availabilities = {}  # Dict for storing available IPs in every subnet
        single_subnet_prov = []  # List for storing provided single subnets
        mult_subnets_prov = []  # List for storing lists of provided multiple subnets

        # For all hosts, get subnets in dict (with no repetitions) and list of
        # provided single & multiple subnets
        for host in hosts:
            subnet_ids = host.get("subnet_ids")
            if subnet_ids:
                if len(subnet_ids) == 1:
                    single_subnet_prov.append(subnet_ids[0])
                    if subnet_ids[0] not in ip_availabilities:
                        ip_availabilities[subnet_ids[0]] = 0
                else:
                    mult_subnets_prov.append(subnet_ids)
                    for subnet_id in subnet_ids:
                        if subnet_id not in ip_availabilities:
                            ip_availabilities[subnet_id] = 0
            else:
                logger.debug(
                    f"{log_msg_start} No subnet/s specified for host {host['name']}."
                )

        # Get available IPs from AWS for every subnet
        logger.info(f"{log_msg_start} Checking IP availability")
        ips_count_wait = [
            self.get_subnet_available_ips(subnet_id, log_msg_start)
            for subnet_id in ip_availabilities
        ]
        ips_count_result = await asyncio.gather(*ips_count_wait)
        ip_availabilities = dict(zip(ip_availabilities, ips_count_result))

        self.subnets_capacity = ip_availabilities.copy()

        # Discount hosts with single subnet specified from available IPs
        for subnet_id in single_subnet_prov:
            if ip_availabilities[subnet_id] > 0:
                ip_availabilities[subnet_id] -= 1
            else:
                logger.info(
                    f"{log_msg_start} Not enougn IP addresses available "
                    f"in subnet: {subnet_id}"
                )
                return False

        # Discount hosts with multiple subnets specified from available IPs
        for subnet_ids in mult_subnets_prov:
            for subnet_id in subnet_ids:
                if ip_availabilities[subnet_id] > 0:
                    ip_availabilities[subnet_id] -= 1
                    break
            else:
                logger.info(
                    f"{log_msg_start} Not enougn IP addresses available "
                    f"in subnet/s: {subnet_ids}"
                )
                return False

        return True

    async def utilization(self):
        """Check percentage utilization of given provider."""
        net_sizes = {
            "28": 14,
            "27": 30,
            "26": 62,
            "25": 126,
            "24": 254,
            "23": 510,
        }
        res = 0
        for net, _availability in self.subnets_capacity.items():
            subnet = self.ec2.Subnet(net)
            size = net_sizes[subnet.cidr_block.split("/")[-1]]
            self.subnets_capacity[net] = subnet.available_ip_address_count
            usage = (size - self.subnets_capacity[net]) / size * 100
            res = usage if usage > res else res

        return res

    async def create_server(self, req):
        """Issue creation of a server.

        req - dict of server requirements

        The req object can contain following additional attributes:
        * 'image': ami or name of image
        * 'flavor': flavor to use

        Returns:
            A tuple containing, respectively, a string (<aws machine id>)
            and a dict (<requirements for the VM>)
            :rtype: (str, dict)
        """
        log_msg_start = f"{self.dsp_name} [{req['name']}]"
        logger.info(f"{log_msg_start} Creating server")
        specs = deepcopy(req)  # work with own copy, do not modify the input

        del_vol = specs.get("delete_volume_on_termination", True)

        name = req.get("name")
        # creating unique name for instance (visible in aws ec2 WebUI)
        taglist = [{"Key": "Name", "Value": name}]
        taglist.append(
            {
                "Key": "Hostname",
                "Value": f"{name.split('.')[0]}-{secrets.token_hex()[:6]}",
            }
        )

        for key, value in self.instance_tags.items():
            taglist.append({"Key": key, "Value": value})

        if specs.get("metadata"):
            for key, value in specs.get("metadata").items():
                taglist.append({"Key": key, "Value": value})

        logger.debug(f"{log_msg_start} Tagging instance with: {object2json(taglist)}")

        request = {
            "ImageId": self.get_image(specs).image_id,
            "MinCount": 1,
            "MaxCount": 1,
            "InstanceType": specs.get("flavor"),
            "KeyName": self.ssh_key,
            "SecurityGroupIds": specs.get("security_group_ids", []),
            "BlockDeviceMappings": [
                {
                    "DeviceName": "/dev/sda1",
                    "Ebs": {
                        "DeleteOnTermination": del_vol,
                    },
                },
            ],
            "TagSpecifications": [{"ResourceType": "instance", "Tags": taglist}],
        }

        subnet_ids = specs.get("subnet_ids")
        if subnet_ids:
            shuffle(subnet_ids)  # Randomize subnets order
            for subnet_id in subnet_ids:
                if self.subnets_capacity[subnet_id] > 0:
                    self.subnets_capacity[subnet_id] -= 1
                    request["SubnetId"] = subnet_id
                    break
            if not request["SubnetId"]:
                raise ProvisioningError(
                    f"There are no subnets with IPs available"
                    f"for use from {subnet_ids}",
                    req,
                )
        else:
            logger.warning(f"{log_msg_start} No subnet/s specified. Using default...")

        if specs.get("spot"):
            request["InstanceMarketOptions"] = {
                "MarketType": "spot",
            }

        try:
            aws_res = self.ec2.create_instances(**request)
        except ClientError as creation_error:
            err_msg = (
                f"{log_msg_start} Requested image "
                f"'{specs.get('image')}' can not be provisioned"
            )
            logger.error(err_msg)
            err_resp = creation_error.response["Error"]["Message"]
            raise ProvisioningError(
                f"{err_msg} Request failed with: {err_resp}", req
            ) from creation_error

        ids = [srv.id for srv in aws_res]
        if len(ids) != 1:  # ids must be len of 1 as we provision one vm at the time
            raise ProvisioningError("Unexpected number of instances provisioned.", req)

        # returns id of provisioned instance and required host name
        return (ids[0], req)

    def get_ip_addresses(self, prov_result):
        """Get IP address from a provisioning result."""
        addresses = []
        if prov_result.get("PublicIpAddress"):
            addresses.append(prov_result.get("PublicIpAddress"))
        if prov_result.get("PrivateIpAddress"):
            addresses.append(prov_result.get("PrivateIpAddress"))
        return addresses

    def prov_result_to_host_data(self, prov_result, req):
        """Transform provisioning result to needed host data."""
        # init the dict
        result = {}

        result["id"] = prov_result.get("InstanceId")
        for tag in prov_result.get("Tags"):
            if tag["Key"] == "Name":
                result["name"] = tag["Value"]  # should be one key "name"

        result["addresses"] = self.get_ip_addresses(prov_result)
        result["status"] = prov_result["State"]["Name"]
        result["os"] = prov_result.get("mrack_req").get("os")
        result["group"] = prov_result.get("mrack_req").get("group")

        return result

    async def wait_till_provisioned(self, resource):
        """Wait for AWS provisioning result."""
        aws_id, req = resource
        instance = self.ec2.Instance(aws_id)
        instance.wait_until_running()
        response = self.client.describe_instances(InstanceIds=[aws_id])
        result = {}
        try:  # returns dict with aws instance information
            result = response["Reservations"][0]["Instances"][0]
            result.update({"mrack_req": req})
        except (KeyError, IndexError) as data_err:
            raise ProvisioningError(
                "Unexpected data format in response "
                f"of provisioned instance '{req['name']}'"
            ) from data_err

        return result, req

    async def delete_host(self, host_id, host_name):
        """Delete provisioned hosts based on input from provision_hosts."""
        log_msg_start = f"{self.dsp_name} [{host_name}]"
        if not host_id:
            logger.debug(
                f"{log_msg_start} Skipping termination, because host was not created"
            )
            return False

        logger.info(f"{log_msg_start} Terminating host with ID {host_id}")
        try:
            self.ec2.instances.filter(InstanceIds=[host_id]).terminate()
        except ClientError as error:
            logger.error(f"{log_msg_start} Issue while terminating host {host_id}:")
            logger.error(error.response["Error"]["Message"])
            return False
        return True



# Copyright 2020 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Ansible inventory output module."""
import logging
import os
import typing
from copy import deepcopy

from mrack.errors import ConfigError
from mrack.outputs.utils import get_external_id
from mrack.utils import (
    get_fqdn,
    get_host_from_metadata,
    get_os_type,
    get_password,
    get_shortname,
    get_ssh_key,
    get_username,
    is_windows_host,
    save_yaml,
)

DEFAULT_INVENTORY_PATH = "mrack-inventory.yaml"
DEFAULT_INVENTORY_LAYOUT: typing.Dict[str, typing.Dict] = {
    "all": {"children": {}, "hosts": {}}
}

logger = logging.getLogger(__name__)


def copy_meta_attrs(host, meta_host, attrs):
    """Copy attributes from host metadata entry and prefixing them with meta."""
    for attr in attrs:
        val = meta_host.get(attr)
        if val:
            host[f"meta_{attr}"] = val


def ensure_all_group(inventory):
    """Ensure that inventory has group "all" with both "hosts" and "children"."""
    if "all" not in inventory:
        inventory["all"] = {
            "children": {},
            "hosts": {},
        }
    all_group = inventory["all"]
    if "children" not in all_group:
        all_group["children"] = {}
    if "hosts" not in all_group:
        all_group["hosts"] = {}
    return all_group


def get_group(inventory, groupname):
    """Get group from inventory, return group or None."""
    groups = inventory.keys()
    found = None
    for g_name in groups:
        group = inventory[g_name]
        if g_name == groupname:
            found = group
        else:
            if "children" in group:
                found = get_group(group["children"], groupname)
        if found is not None:
            break
    return found


def add_to_group(inventory, groupname, hostname):
    """
    Find a group in inventory layout and add a host to it.

    Returns group or None if it doesn't exist.
    """
    group = get_group(inventory, groupname)
    if group is None:
        return None

    if "hosts" not in group:
        group["hosts"] = {}
    group["hosts"][hostname] = {}
    return group


def add_group(inventory, groupname, hostname=None):
    """Add a group to inventory, optionally add it with hostname."""
    all_group = inventory["all"]
    group = {}
    if hostname:
        group["hosts"] = {hostname: {}}
    all_group["children"][groupname] = group
    return group


class AnsibleInventoryOutput:
    """
    Generate Ansible inventory with provisioned machines.

    The resulting inventory combines data from provisioning config, job metadata
    files and information in DB.
    """

    def __init__(self, config, db, metadata, path=None):  # pylint: disable=invalid-name
        """Init the output module."""
        self._config = config
        self._db = db
        self._metadata = metadata
        self._path = path or DEFAULT_INVENTORY_PATH

    def create_ansible_host(self, name):
        """Create host entry for Ansible inventory."""
        # pylint: disable=too-many-locals
        meta_host, meta_domain = get_host_from_metadata(self._metadata, name)
        db_host = self._db.hosts[name]

        ip_addr = db_host.ip_addr
        ansible_host = get_external_id(db_host, meta_host, self._config)

        python = (
            self._config["python"].get(meta_host["os"])
            or self._config["python"]["default"]
        )

        ansible_user = get_username(db_host, meta_host, self._config)
        password = get_password(db_host, meta_host, self._config)
        ssh_key = get_ssh_key(db_host, meta_host, self._config)
        dom_name = meta_domain["name"]

        # Common attributes
        dc_list = [f"DC={dc}" for dc in dom_name.split(".")]
        host_info = {
            "ansible_host": ansible_host,
            "ansible_python_interpreter": python,
            "ansible_user": ansible_user,
            "meta_os_type": get_os_type(meta_host),
            "meta_fqdn": f"{get_fqdn(name, dom_name)}",
            "meta_hostname": get_shortname(name),
            "meta_domain": f"{get_fqdn(name, dom_name).split('.', 1)[1]}",
            "meta_provider": db_host.provider.name,
            "meta_provider_id": db_host.host_id,
            "meta_ip": ip_addr,
            "meta_dc_record": ",".join(dc_list),
        }

        if "restraint_id" in meta_host:
            host_info.update({"meta_restraint_id": meta_host["restraint_id"]})

        copy_meta_attrs(host_info, meta_host, ["os", "role", "netbios"])

        if "parent" in meta_domain:
            host_info["meta_parent_domain"] = meta_domain["parent"]

        if ssh_key:
            host_info["ansible_ssh_private_key_file"] = os.path.abspath(ssh_key)

        if password:
            host_info["ansible_password"] = password

        if db_host.provider.name in ("docker", "podman"):
            host_info.update({"ansible_user": "root"})  # TODO make it configurable

        if db_host.meta_extra:
            for key, value in db_host.meta_extra.items():
                host_info.update({key: value})

        if is_windows_host(meta_host):
            if "netbios" in meta_host:
                host_info.update({"meta_netbios": meta_host["netbios"]})

            host_info.update(
                {
                    "ansible_port": 5986,
                    "ansible_connection": "winrm",
                    "ansible_winrm_server_cert_validation": "ignore",
                    "meta_domain_level": meta_host.get("domain_level", "top"),
                }
            )

        for key, val in meta_host.items():
            if key.startswith("meta_"):
                host_info[key] = val

        glob_ansible_inv = deepcopy(self._metadata.get("ansible_inventory", {}))
        dom_ansible_inv = deepcopy(meta_domain.get("ansible_inventory", {}))
        host_ansible_inv = deepcopy(meta_host.get("ansible_inventory", {}))

        if any([glob_ansible_inv, dom_ansible_inv, host_ansible_inv]):
            host_info.update(glob_ansible_inv | dom_ansible_inv | host_ansible_inv)

        return host_info

    def get_inventory_layout(self):
        """Get Ansible inventory layout."""
        layout = self._metadata.get("config", {}).get("ansible", {}).get("layout")
        if layout is None:
            return self._config.get("inventory_layout", DEFAULT_INVENTORY_LAYOUT)

        return layout

    def create_inventory(self):
        """Create the Ansible inventory in dict form."""
        provisioned = self._db.hosts
        inventory = deepcopy(self.get_inventory_layout())
        if not isinstance(inventory, dict):
            raise ConfigError("Inventory layout should be a dictionary")
        all_group = ensure_all_group(inventory)

        for host in provisioned.values():
            meta_host, _meta_domain = get_host_from_metadata(self._metadata, host.name)
            if meta_host is None:
                continue

            # Groups can be defined in both "groups" and "group" variable.
            groups = meta_host.get("groups", [])
            group = meta_host.get("group")
            if group and group not in groups:
                groups.append(group)

            # Add only a reference custom groups
            for group in groups:
                added = add_to_group(inventory, group, host.name)
                if not added:  # group doesn't exist
                    add_group(inventory, group, host.name)

            # Main record belongs in "all" group
            all_group["hosts"][host.name] = self.create_ansible_host(host.name)
        return inventory

    def create_output(self):
        """Create the target output file."""
        inventory = self.create_inventory()
        save_yaml(self._path, inventory)
        if self._path:
            logger.info(f"Created: {self._path}")
        return inventory



# Copyright 2020 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Output action."""
import logging

from mrack.actions.action import Action
from mrack.context import global_context
from mrack.errors import MetadataError
from mrack.outputs.ansible_inventory import AnsibleInventoryOutput
from mrack.outputs.pytest_mh import PytestMhOutput
from mrack.outputs.pytest_multihost import PytestMultihostOutput

logger = logging.getLogger(__name__)


class Output(Action):
    """
    Output action.

    Test action to run output modules from data in database.
    """

    def __init__(
        self,
        config=None,
        metadata=None,
        db_driver=None,
        ansible_path=None,
        pytest_multihost_path=None,
        pytest_mh_path=None,
    ):  # pylint: disable=arguments-differ
        """Initialize the Output action."""
        super().__init__(config=config, metadata=metadata, db_driver=db_driver)

        if global_context.CONFIG:
            global_ansible_path = global_context.CONFIG.ansible_inventory_path()
            global_multihost_path = global_context.CONFIG.pytest_multihost_path()
            global_mh_path = global_context.CONFIG.pytest_mh_path()
        else:
            global_ansible_path = None
            global_multihost_path = None
            global_mh_path = None

        self._ansible_path = ansible_path or global_ansible_path
        self._pytest_multihost_path = pytest_multihost_path or global_multihost_path
        self._pytest_mh_path = pytest_mh_path or global_mh_path

    async def generate_outputs(self):
        """Generate outputs."""
        logger.info("Output generation started")

        outputs = self._metadata.get("config", {}).get(
            "outputs", ["ansible-inventory", "pytest-multihost"]
        )
        outputs_map = {
            "ansible-inventory": (AnsibleInventoryOutput, self._ansible_path),
            "pytest-multihost": (PytestMultihostOutput, self._pytest_multihost_path),
            "pytest-mh": (PytestMhOutput, self._pytest_mh_path),
        }

        if not self._db_driver.hosts:
            raise MetadataError("No hosts found.")

        logger.info("Requested outputs: " + ", ".join(outputs))

        for output in outputs:
            (cls, path) = outputs_map[output]
            o = cls(self._config, self._db_driver, self._metadata, path)
            o.create_output()

        logger.info("Output generation done")
        return True



# Copyright 2020 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""General Provider interface."""
import asyncio
import logging
import random
import socket
from datetime import datetime, timedelta

from mrack.context import global_context
from mrack.errors import ProvisioningError
from mrack.host import STATUS_ACTIVE, STATUS_OTHER, Host
from mrack.utils import get_username_pass_and_ssh_key, object2json, ssh_to_host

logger = logging.getLogger(__name__)

STRATEGY_ABORT = "abort"
STRATEGY_RETRY = "retry"
RET_CODE = 0  # index to access return code from _wait_for_ssh
HOST_OBJ = 1  # index to access host object from _wait_for_ssh
ERROR_OBJ = 0  # default index to access host error which caused ProvisioningError
SPECS = 1  # default index to access host specs which caused ProvisioningError


class Provider:
    """General Provider interface."""

    def __init__(self):
        """Initialize provider."""
        self._name = "dummy"
        self.dsp_name = "Dummy"
        self.timeout = 60
        self.max_retry = 1
        self.strategy = STRATEGY_ABORT
        self.status_map = {"OTHER": STATUS_OTHER}

    @property
    def name(self):
        """Get provider name."""
        return self._name

    async def validate_hosts(self, reqs):
        """Validate that host requirements are well specified."""
        raise NotImplementedError()

    async def can_provision(self, hosts):
        """Check that provider has enough resources to provision hosts."""
        raise NotImplementedError()

    async def utilization(self):
        """Check percentage utilization of given provider."""
        raise NotImplementedError()

    async def create_server(self, req):
        """Request and create resource on selected provider."""
        raise NotImplementedError()

    async def wait_till_provisioned(self, resource):
        """Wait till resource is provisioned."""
        raise NotImplementedError()

    async def prepare_provisioning(self, reqs):
        """Prepare provisioning."""
        raise NotImplementedError()

    async def _wait_for_ssh(self, host, timeout, port):
        """
        Wait until a port starts accepting TCP connections and is able to connect.

        Args:
            host (Host): Host object to get its address on which the port should exist.
            timeout (float): In minutes. How long to wait before raising errors.
            port (int): Port number.
        Raises:
            TimeoutError: The port isn't accepting connection after specified `timeout`.
        """
        log_msg_start = f"{self.dsp_name} [{host.name}]"
        start_time = datetime.now()
        info_msg = (
            f"{log_msg_start} Waiting for the port {port} on host "
            f"{host.ip_addr} to start accepting connections (up to {timeout} minutes)"
        )
        logger.info(info_msg)

        while True:
            try:
                with socket.create_connection(
                    (host.ip_addr, port), timeout=(timeout * 60)
                ):
                    logger.info(
                        f"{log_msg_start} Port {port} on host "
                        f" {host.ip_addr} is now open"
                    )
                    break
            except OSError:
                await asyncio.sleep(10)
                logger.debug(info_msg)
                if datetime.now() - start_time >= timedelta(seconds=(timeout * 60)):
                    logger.error(
                        f"{log_msg_start} Waited too long for the port "
                        f"{port} on host {host.ip_addr} to start accepting connections"
                    )
                    # do not continue to try ssh connection after port is not open
                    return False, host

        # Wait also for the ssh key to be accepted for a half timeout time
        start_ssh = datetime.now()
        # load stuff from configs: like:
        username, password, ssh_key = get_username_pass_and_ssh_key(
            host, global_context
        )

        info_msg = (
            f"{log_msg_start} Waiting for the host {host.ip_addr} "
            f"to accept the authentication method (up to {timeout} minutes)"
        )

        while True:
            res = ssh_to_host(
                host,
                username=username,
                password=password,
                ssh_key=ssh_key,
                command="echo mrack",
            )
            duration = (datetime.now() - start_ssh).total_seconds()

            if res:
                logger.info(
                    f"{log_msg_start} SSH to host '{host.ip_addr}' successful "
                    f"after {duration:.1f}s"
                )
                break

            if datetime.now() - start_ssh >= timedelta(seconds=(timeout * 60)):
                logger.error(
                    f"{log_msg_start} SSH to host '{host.ip_addr}' "
                    f"timed out after {duration:.1f}s"
                )
                break

            # wait 10 seconds to retry the ssh connection
            await asyncio.sleep(10)

        return res, host

    async def _check_ssh_auth(self, default_check, active_hosts):
        """Check the ssh authentication functionality."""
        success_hosts = []
        error_hosts = []

        if not isinstance(default_check, dict):
            default_check = {}

        # check if default config is in place, if not use defaults
        req_keys = ("enabled", "port", "timeout")
        if not all(k in default_check for k in req_keys):
            logger.warning(
                f"{self.dsp_name} Missing complete default post provisioning "
                "ssh check configuration in provisioning config file."
            )
            # extend not complete configuration with defaults
            default_check = {
                "enabled": True,
                "port": 22,
                "timeout": 10,
            } | default_check

        # split dictionary into two default and based on host os/group
        req_keys += ("enabled_providers", "disabled_providers")
        based_check = {x: default_check[x] for x in default_check if x not in req_keys}
        default_check = {x: default_check[x] for x in default_check if x in req_keys}

        wait_ssh = []
        for host in active_hosts:
            # go host by host and load the group and os based configuration for ssh
            os_check = based_check.get("os", {}).get(host.operating_system, {})
            group_check = based_check.get("group", {}).get(host.group, {})
            opts = default_check | group_check | os_check  # sorted by priority

            logger.debug(
                f"{self.dsp_name} [{host.name}] "
                f"ssh check config: {object2json(opts)}"
            )

            if (
                not opts.get("enabled")
                and self.name not in opts.get("enabled_providers", [])
            ) or (
                opts.get("enabled") and self.name in opts.get("disabled_providers", [])
            ):
                success_hosts.append(host)
                logger.debug(
                    f"{self.dsp_name} Skipping ssh check for host '{host.name}'"
                )
                continue

            awaitable = self._wait_for_ssh(
                host,
                timeout=opts.get("timeout"),
                port=opts.get("port"),
            )
            wait_ssh.append(awaitable)

        ssh_results = await asyncio.gather(*wait_ssh)
        # We distinguish the success hosts and new error hosts from active by using:
        # res[RET_CODE] 0
        #   - the result of operation returned from self._wait_for_ssh()
        # res[HOST_OBJ] 1
        #   - the host object returned from self._wait_for_ssh()
        for res in ssh_results:
            if res[RET_CODE]:
                success_hosts.append(res[HOST_OBJ])
            else:
                res[HOST_OBJ].error = (
                    "Could not establish ssh connection to host "
                    f"{res[HOST_OBJ].host_id} with IP {res[HOST_OBJ].ip_addr}"
                )
                error_hosts.append(res[HOST_OBJ])

        return success_hosts, error_hosts

    async def _provision_base(
        self,
        reqs,
        timeout=None,
    ):  # pylint: disable=too-many-locals, too-many-branches, too-many-statements
        """Provision hosts based on list of host requirements.

        Main function which does provisioning and validation.
        Parameters:
            reqs - dictionary with requirements for provider
            timeout - base timeout (minutes) to wait for resources
        """
        if not timeout:
            timeout = self.timeout

        log_msg_start = f"{self.dsp_name}"  # use display name at log message beginning
        logger.info(f"{log_msg_start} Validating host(s) definitions")
        if not reqs:
            raise ProvisioningError(
                f"{log_msg_start} Can not continue with empty requirement for provider"
            )

        await self.validate_hosts(reqs)
        logger.info(f"{log_msg_start} Host(s) definitions valid")

        delta = abs(self.timeout - timeout)
        slow_down = random.choice(range(delta, self.timeout + delta))

        logger.debug(
            f"{log_msg_start} Slowing down the execution for {slow_down} s "
            "(This will help mrack to get better resource overview for concurrent runs)"
        )

        await asyncio.sleep(slow_down)
        # This slows down the provisioning at the beginning using a variable time
        # to sleep so we prevent mrack executions to provision at the same time.
        logger.info(
            f"{log_msg_start} Setting timeout to wait "
            f"for resources to {timeout} min(s)"
        )

        logger.info(f"{log_msg_start} Checking available resources")
        res_check_start = datetime.now()
        error_hosts = []

        while not await self.can_provision(reqs):
            if datetime.now() - res_check_start >= timedelta(minutes=timeout):
                # create error host object so retry strategy can continue
                # instead of throwing exception to fail at once without retry
                err_str = "Not enough resources to provision"
                logger.error(f"{log_msg_start} {err_str}")
                for req in reqs:
                    error_hosts.append(
                        Host(
                            provider=self,
                            host_id=req.get("name"),
                            name=req.get("name"),
                            operating_system=req.get("os"),
                            group=req.get("group"),
                            ip_addrs=[],
                            status=STATUS_OTHER,
                            rawdata=req,
                            error_obj=err_str,
                        )
                    )
                return ([], error_hosts, reqs)

            logger.info(
                f"{log_msg_start} Not enough resources to provision, "
                f"checking again in {timeout * 10} seconds(s)"
            )
            # Sleep time to wait to check if resources are available again.
            await asyncio.sleep(timeout * 10)

        logger.info(f"{log_msg_start} Resource availability: OK")
        started = datetime.now()

        logger.info(f"{log_msg_start} Issuing provisioning of {len(reqs)} host(s)")
        create_servers = []
        for req in reqs:
            awaitable = self.create_server(req)
            create_servers.append(awaitable)

        # expect the exception in return data to be parsed later
        create_resps = await asyncio.gather(*create_servers, return_exceptions=True)

        logger.info(f"{log_msg_start} Provisioning issued")

        logger.info(f"{log_msg_start} Waiting for all hosts to be active")

        wait_servers = []
        for response in create_resps:
            if isinstance(response, ProvisioningError):
                # use ProvisioningError arguments to create missing Host object
                # which we append to error hosts list for later usage
                error_hosts.append(
                    Host(
                        provider=self,
                        host_id=response.args[SPECS].get("host_id"),
                        name=response.args[SPECS].get("name"),
                        operating_system=response.args[SPECS].get("os"),
                        group=response.args[SPECS].get("group"),
                        ip_addrs=[],
                        status=STATUS_OTHER,
                        rawdata=response.args,
                        error_obj=response.args[ERROR_OBJ],
                    )
                )
            elif isinstance(response, Exception):
                logger.error("An unexpected exception occurred while provisioning")
                raise response
            else:
                # response might be okay so let us wait for result
                awaitable = self.wait_till_provisioned(response)
                wait_servers.append(awaitable)

        server_results = await asyncio.gather(*wait_servers)
        provisioned = datetime.now()

        logger.info(
            f"{log_msg_start} "
            "All hosts reached provisioning final state (ACTIVE or ERROR)"
        )
        logger.info(f"{log_msg_start} Provisioning duration: {provisioned - started}")

        hosts = [self.to_host(srv, req) for srv, req in server_results if srv]

        error_hosts += await self.parse_error_hosts(hosts)
        active_hosts = [h for h in hosts if h not in error_hosts]
        success_hosts = []

        ssh_check = global_context.PROV_CONFIG.get("post_provisioning_check", {}).get(
            "ssh", True
        )  # enable check by default

        # check ssh connectivity to hosts if not disabled per host or provider
        if bool(ssh_check):
            success_hosts, failed_hosts = await self._check_ssh_auth(
                ssh_check, active_hosts
            )
            error_hosts += failed_hosts
        else:  # we do not check the ssh connection to VMs
            success_hosts = active_hosts

        missing_reqs = [
            req for req in reqs if req["name"] in [host.name for host in error_hosts]
        ]

        return (success_hosts, error_hosts, missing_reqs)

    async def provision_hosts(self, reqs):
        """Provision hosts based on list of host requirements.

        Main provider method for provisioning.

        Issues provisioning and waits for it succeed. Raises exception if any of
        the servers was not successfully provisioned. If that happens it issues deletion
        of all already provisioned resources.

        Return list of information about provisioned servers.
        """
        log_msg_start = self.dsp_name
        logger.info(f"{log_msg_start} Preparing provider resources")
        if not await self.prepare_provisioning(reqs):
            raise ProvisioningError("Failed to prepare resources", self.dsp_name)

        if self.strategy == STRATEGY_RETRY:
            success_hosts, error_hosts, _missing_reqs = await self.strategy_retry(reqs)
        else:
            success_hosts, error_hosts, _missing_reqs = await self.strategy_abort(reqs)

        if error_hosts:
            hosts_to_delete = success_hosts + error_hosts
            await self.abort_and_delete(hosts_to_delete, error_hosts)

        logger.info(f"{log_msg_start} Printing provisioned hosts")
        for host in success_hosts:
            logger.info(f"{log_msg_start} {host}")

        return success_hosts

    async def strategy_retry(self, reqs):
        """Provisioning strategy to try multiple times to provision a host."""
        log_msg_start = self.dsp_name
        missing_reqs = reqs
        attempts = 0
        success_hosts = []
        error_hosts = []
        delta_sleep = global_context.CONFIG.delta_sleep
        max_utilization = global_context.CONFIG.max_utilization

        while missing_reqs:
            # number of attempts should be greater than max_retry
            # and not greater or equal than max_retry because
            # other way first attempt to provision will be
            # considered as first retry and provisioning will end
            # after reaching count 1 without reprovisioning retry
            if attempts > self.max_retry:
                logger.error(
                    f"{log_msg_start} Max retry attempts "
                    f"({self.max_retry}) reached. Aborting"
                )
                break

            attempts += 1

            # set the waiting timeout to vary from 45-75 minutes randomly
            # so if there are multiple parallel runs of mrack they wait
            # different amount time before poll and increase chance to get resources
            # quicker than other concurrent runs of mrack requests
            delta = random.choice(range(-delta_sleep, delta_sleep))
            res_check_timeout = self.timeout + delta

            s_hosts, error_hosts, missing_reqs = await self._provision_base(
                missing_reqs, timeout=res_check_timeout
            )
            success_hosts.extend(s_hosts)

            # We need to cleanup the error hosts for the next retry run
            # in case of last attempt which is `self.max_retry`
            # we skip this part at it done in provision_hosts() method while
            # awaiting the self.abort_and_delete(error hosts)

            if error_hosts and attempts <= self.max_retry:
                count = len(error_hosts)
                err = f"{count} hosts were not provisioned properly, deleting."
                logger.info(f"{log_msg_start} {err}")
                out_of_resources = False
                for host in error_hosts:
                    logger.error(f"{log_msg_start} Error: {str(host.error)}")
                    # if host.error is a dictionary and code is 500 = SERVER ERROR
                    if (
                        isinstance(host.error, dict)
                        and host.error.get("code", None) == 500
                    ):
                        logger.info(
                            f"{log_msg_start} Provider is out of resources, "
                            "increasing cooldown time before another retry"
                        )
                        out_of_resources = True

                if await self.utilization() >= max_utilization or out_of_resources:
                    # delete all hosts when there is high utilization of provider
                    error_hosts += success_hosts

                await self.delete_hosts(error_hosts)

                # multiply the sleep time to increase the cooldown period
                # to get range from 15 to 30 minutes
                # before another retry (using default delta_sleep=15)
                res_check_timeout += (
                    (delta_sleep - 1) * res_check_timeout * int(out_of_resources)
                )  # -1 because we do addition to original value and multiply by 14

                cooldown = random.choice(
                    range(res_check_timeout, 2 * res_check_timeout)
                )
                logger.info(
                    f"{log_msg_start} Retrying to provision these hosts in {cooldown}s"
                )
                # In this case we haven't been able to provision all resources
                # despite the previous waits, so we probably are in a race condition
                # with concurrent runs, and thus we free all resources and try again
                # in `cooldown` seconds to give chance other mrack requests.
                await asyncio.sleep(cooldown)

        return success_hosts, error_hosts, missing_reqs

    async def strategy_abort(self, reqs):
        """Provisioning strategy to try once and then abort."""
        return await self._provision_base(reqs)

    async def parse_error_hosts(self, hosts):
        """Parse provisioning errors from provider result."""
        log_msg_start = self.dsp_name
        errors = []
        logger.debug(f"{log_msg_start} Checking provisioned hosts for errors")
        for host in hosts:
            logger.debug(
                f"{log_msg_start} [{host.name}] ID {host.host_id}"
                f"\tSTATUS - {host.status}"
            )

            if host.status != STATUS_ACTIVE:
                errors.append(host)

        return errors

    async def abort_and_delete(self, hosts_to_delete, error_hosts):
        """Delete hosts and abort provisioning with an error."""
        log_msg_start = self.dsp_name
        logger.info(f"{log_msg_start} Aborting provisioning due to error.")
        for host in error_hosts:
            logger.error(f"{log_msg_start} [{host.name}] Error: {str(host.error)}")

        logger.info(f"{log_msg_start} Given the error, will delete hosts")
        await self.delete_hosts(hosts_to_delete)
        raise ProvisioningError(
            f"{log_msg_start} Failed to provision {len(error_hosts)} host(s) "
            "due to provisioning error",
            self.dsp_name,
        )

    async def delete_host(self, host_id, host_name):
        """Delete provisioned host."""
        raise NotImplementedError()

    async def delete_hosts(self, hosts):
        """Issue deletion of all servers based on previous results from provisioning."""
        log_msg_start = self.dsp_name
        logger.info(f"{log_msg_start} Issuing deletion")
        delete_servers = []
        for host in hosts:
            awaitable = self.delete_host(host.host_id, host.name)
            delete_servers.append(awaitable)
        results = await asyncio.gather(*delete_servers)
        logger.info(f"{log_msg_start} All servers issued to be deleted")
        return results

    def prov_result_to_host_data(self, prov_result, req):
        """Transform provisioning result to needed host data."""
        raise NotImplementedError()

    def to_host(self, provisioning_result, req, username=None):
        """Transform provisioning result into Host object."""
        host_info = self.prov_result_to_host_data(provisioning_result, req)

        host = Host(
            self,
            host_info.get("id"),
            host_info.get("name"),
            host_info.get("os"),
            host_info.get("group"),
            host_info.get("addresses"),
            self.status_map.get(host_info.get("status"), STATUS_OTHER),
            provisioning_result,
            username=username,
            password=host_info.get("password"),
            error_obj=host_info.get("fault"),
            meta_extra=host_info.get("meta_extra"),
        )
        return host



# Copyright 2020 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""mrack default app."""
#  pylint: disable=no-name-in-module
import logging
import sys
from typing import Set, Tuple, Type

import click

from mrack.actions.destroy import Destroy
from mrack.actions.etchosts import EtcHostsUpdate
from mrack.actions.list import List
from mrack.actions.output import Output
from mrack.actions.ssh import SSH
from mrack.actions.up import Up
from mrack.context import GlobalContext, global_context
from mrack.errors import ApplicationError, MrackError
from mrack.providers import providers
from mrack.providers.provider import Provider
from mrack.utils import async_run
from mrack.version import VERSION

PROVIDER_NAME = 0
PROVIDER_CLASS = 1
installed_providers: Set[Tuple[str, Type[Provider]]] = set()
logger = logging.getLogger(__name__)
IMPORT_ERR_TEMPLATE = "Provider '%s' not installed, skipping provider registration"


try:
    from mrack.providers.aws import PROVISIONER_KEY as AWS
    from mrack.providers.aws import AWSProvider

    installed_providers.add((AWS, AWSProvider))
except ModuleNotFoundError as err:
    logger.debug(IMPORT_ERR_TEMPLATE, err.name)

try:
    from mrack.providers.beaker import PROVISIONER_KEY as BEAKER
    from mrack.providers.beaker import BeakerProvider

    installed_providers.add((BEAKER, BeakerProvider))
except ModuleNotFoundError as err:
    logger.debug(IMPORT_ERR_TEMPLATE, err.name)

try:
    from mrack.providers.openstack import PROVISIONER_KEY as OPENSTACK
    from mrack.providers.openstack import OpenStackProvider

    installed_providers.add((OPENSTACK, OpenStackProvider))
except ModuleNotFoundError as err:
    logger.debug(IMPORT_ERR_TEMPLATE, err.name)

try:
    from mrack.providers.podman import PROVISIONER_KEY as PODMAN
    from mrack.providers.podman import PodmanProvider

    installed_providers.add((PODMAN, PodmanProvider))
except ModuleNotFoundError as err:
    logger.debug(IMPORT_ERR_TEMPLATE, err.name)

try:
    from mrack.providers.static import PROVISIONER_KEY as STATIC
    from mrack.providers.static import StaticProvider

    installed_providers.add((STATIC, StaticProvider))
except ModuleNotFoundError as err:
    logger.debug(IMPORT_ERR_TEMPLATE, err.name)

try:
    from mrack.providers.virt import PROVISIONER_KEY as VIRT
    from mrack.providers.virt import VirtProvider

    installed_providers.add((VIRT, VirtProvider))
except ModuleNotFoundError as err:
    logger.debug(IMPORT_ERR_TEMPLATE, err.name)


def init_providers():
    """Register all providers usable in this session."""
    if not installed_providers:
        raise ApplicationError("FATAL: mrack did not find any installed providers")

    for installed_provider in installed_providers:
        providers.register(
            installed_provider[PROVIDER_NAME],
            installed_provider[PROVIDER_CLASS],
        )

    logger.info(f"Installed providers: {', '.join(providers.names)}")


async def generate_outputs(ctx):
    """Init and run output action."""
    config = ctx.obj.CONFIG
    output_action = Output(
        ctx.obj.PROV_CONFIG,
        ctx.obj.METADATA,
        ctx.obj.DB,
        config.ansible_inventory_path(),
        config.pytest_multihost_path(),
        config.pytest_mh_path(),
    )
    await output_action.generate_outputs()


@click.group()
@click.option("-c", "--mrack-config", type=click.Path(exists=True))
@click.option("-p", "--provisioning-config", type=click.Path(exists=True))
@click.option("-d", "--db", "db_file")  # db file may not exist
@click.option("--debug", default=False, is_flag=True)
@click.version_option(version=VERSION)
@click.pass_context
def mrackcli(ctx, mrack_config, provisioning_config, db_file, debug):
    """Multihost human friendly provisioner."""
    if debug:
        logging.getLogger("mrack").setLevel(logging.DEBUG)

    init_providers()
    ctx.ensure_object(GlobalContext)
    global_context.init(mrack_config, provisioning_config, db_file)
    ctx.obj = global_context


@mrackcli.command()
@click.pass_context
@click.option("-m", "--metadata", type=click.Path(exists=True))
@click.option("-p", "--provider", default="openstack")
@async_run
async def up(ctx, metadata, provider):  # pylint: disable=invalid-name
    """Provision hosts.

    Based on provided job metadata file and provisioning configuration.
    """
    ctx.obj.init_metadata(metadata)

    up_action = Up(ctx.obj.PROV_CONFIG, ctx.obj.METADATA, ctx.obj.DB)
    await up_action.init(provider)
    await up_action.provision()

    await generate_outputs(ctx)


@mrackcli.command()
@click.pass_context
@click.option("-m", "--metadata", type=click.Path(exists=True))
@async_run
async def destroy(ctx, metadata):
    """Destroy provisioned hosts."""
    ctx.obj.init_metadata(metadata)
    destroy_action = Destroy(ctx.obj.PROV_CONFIG, ctx.obj.METADATA, ctx.obj.DB)
    await destroy_action.destroy()


@mrackcli.command()
@click.pass_context
@click.option("-m", "--metadata", type=click.Path(exists=True))
@async_run
async def output(ctx, metadata):
    """Create outputs - such as Ansible inventory."""
    ctx.obj.init_metadata(metadata)
    await generate_outputs(ctx)


@mrackcli.command()
@click.pass_context
@async_run
async def list(ctx):  # pylint: disable=redefined-builtin
    """List host tracked by."""
    list_action = List(ctx.obj.DB)
    list_action.list()


@mrackcli.command()
@click.pass_context
@click.argument("hostname", required=False)
@click.option("-m", "--metadata", type=click.Path(exists=True))
@async_run
async def ssh(ctx, hostname, metadata):
    """SSH to host."""
    ctx.obj.init_metadata(metadata)
    ssh_action = SSH(ctx.obj.PROV_CONFIG, ctx.obj.METADATA, ctx.obj.DB)
    return ssh_action.ssh(hostname)


@mrackcli.group()
def eh():  # pylint: disable=invalid-name
    """Commands to update /etc/hosts file."""


@eh.command()
@click.pass_context
def add(ctx):
    """Add active hosts to /etc/hosts file."""
    eh_action = EtcHostsUpdate(ctx.obj.DB)
    eh_action.update()


@eh.command()
@click.pass_context
def clear(ctx):
    """Remove all mrack hosts from /etc/hosts file."""
    eh_action = EtcHostsUpdate(ctx.obj.DB)
    eh_action.clear()


def exception_handler(func):
    """
    Top level exception handler.

    For showing nice output to users if exception bubbles up to the top.
    """

    def handle(*args, **kwargs):
        """Handle exceptions."""
        ret_code = 1  # assuming error
        try:
            ret_code = func(*args, **kwargs)
        except (
            FileNotFoundError,
            MrackError,
        ) as known_error:
            logger.error(known_error)
            sys.exit(1)
        except Exception as exc:
            logger.exception(exc)
            raise exc

        return ret_code

    return handle


@exception_handler
def run():
    """Run the app."""
    logger.debug(f"mrack version: {VERSION}")
    mrackcli(obj={})  # pylint: disable=no-value-for-parameter,unexpected-keyword-arg


if __name__ == "__main__":
    run()



# Copyright 2020 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Host object."""

from mrack.providers import providers
from mrack.utils import object2json

STATUS_PENDING = "pending"
STATUS_PROVISIONING = "provisioning"
STATUS_ACTIVE = "active"
STATUS_ERROR = "error"
STATUS_DELETED = "deleted"
STATUS_DELETING = "deleting"
STATUS_OTHER = "other"

STATUSES = [
    STATUS_PENDING,
    STATUS_PROVISIONING,
    STATUS_ACTIVE,
    STATUS_ERROR,
    STATUS_DELETED,
    STATUS_DELETING,
]


def host_from_json(host_data):
    """Reverse method to Host.__json__() after json.loads()."""
    provider_name = host_data["provider"]
    provider = providers.get(provider_name)
    host = Host(
        provider,
        host_data["host_id"],
        host_data["name"],
        host_data["operating_system"],
        host_data["group"],
        host_data["ip_addrs"],
        host_data["status"],
        host_data["rawdata"],
        host_data["username"],
        host_data["password"],
        host_data["error"],
        host_data.get("meta_extra"),
    )
    return host


class Host:
    """Provisioned host.

    Normalized values from providers to offer consistent interface.
    """

    def __init__(
        self,
        provider,
        host_id,
        name,
        operating_system,
        group,
        ip_addrs,
        status,
        rawdata,
        username=None,
        password=None,
        error_obj=None,
        meta_extra=None,
    ):
        """Initialize host object."""
        self._provider = provider
        self._host_id = host_id
        self._name = name
        self._operating_system = operating_system
        self._group = group
        self._ip_addrs = ip_addrs
        self._status = status
        self._username = username
        self._password = password
        self._rawdata = rawdata
        self._error = error_obj
        self._meta_extra = meta_extra

    def __str__(self):
        """Return string representation of host."""
        net_str = " ".join(self._ip_addrs)

        out = (
            f"{self._status} {self._operating_system} {self._host_id} {self._name} "
            f"{net_str} {self._username} {self._password}"
        )

        if self._error:
            out = "\n".join(
                [
                    out,
                    f"Error: Host finished with error(s): {object2json(self._error)}",
                ]
            )
        return out

    def to_json(self):
        """Transform object into representation which is acceptable by `json.dump`."""
        return {
            "provider": self._provider.name,
            "host_id": self._host_id,
            "name": self._name,
            "operating_system": self._operating_system,
            "group": self._group,
            "ip_addrs": self._ip_addrs,
            "status": self._status,
            "username": self._username,
            "password": self._password,
            "rawdata": self._rawdata,
            "error": self._error,
            "meta_extra": self._meta_extra,
        }

    @property
    def provider(self):
        """Get host provisioning provider."""
        return self._provider

    @property
    def operating_system(self):
        """Get host operating system."""
        return self._operating_system

    @property
    def group(self):
        """Get host group."""
        return self._group

    @property
    def host_id(self):
        """Get provider host id."""
        return self._host_id

    @property
    def name(self):
        """Get host name."""
        return self._name

    @property
    def ip_addrs(self):
        """Get host IP addresses."""
        return self._ip_addrs

    @property
    def ip_addr(self):
        """Get first host IP address."""
        return self._ip_addrs[0] if self._ip_addrs else ""

    @property
    def status(self):
        """Get host status."""
        return self._status

    @property
    def error(self):
        """Get host error object."""
        return self._error

    @error.setter
    def error(self, value):
        """Set host error object."""
        self._error = value

    @property
    def username(self):
        """Get username for connecting to host."""
        return self._username

    @property
    def password(self):
        """Get password for connecting to host."""
        return self._password

    @property
    def meta_extra(self):
        """Get host extra meta information."""
        return self._meta_extra

    async def delete(self):
        """Issue host deletion via associated provider."""
        await self.provider.delete_host(self.host_id, self.name)
        self._status = STATUS_DELETED
        return True



# Copyright 2023 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""pytest-mh configuration file output module."""

import logging

from mrack.outputs.utils import get_external_id
from mrack.utils import get_fqdn, get_os_type, save_yaml

DEFAULT_MHCFG_PATH = "pytest-mh.yaml"

logger = logging.getLogger(__name__)


class PytestMhOutput:
    """
    Create a configuration file for pytest-mh.

    From job metadata and provisioned information (DB).

    Structure of mh config file is almost the same as our topology in job
    metadata definition.
    """

    def __init__(self, config, db, metadata, path=None):
        """Init the output module."""
        self._config = config
        self._db = db
        self._metadata = metadata
        self._path = path or DEFAULT_MHCFG_PATH

    def create_mh_config(self):
        """
        Create configuration file for python-pytest-multihost.

        Return in form of dict.
        """
        cfg = {"domains": []}
        for domain in self._metadata.get("domains", []):
            cfgdom = {"id": domain["name"], "hosts": []}
            cfg["domains"].append(cfgdom)

            for host in domain.get("hosts", []):
                if "role" not in host:
                    continue

                provisioned_host = self._db.hosts.get(host["name"], None)
                if not provisioned_host:
                    logger.error(f"Host {host['name']} not found in the database.")
                    continue

                cfgdom["hosts"].append(
                    {
                        "hostname": get_fqdn(host["name"], domain.get("name", "")),
                        "os": {
                            "family": get_os_type(host),
                        },
                        "role": host["role"],
                        "ssh": {
                            "host": get_external_id(
                                provisioned_host, host, self._config
                            ),
                        },
                        **host.get("pytest_mh", {}),
                    }
                )

        return cfg

    def create_output(self):
        """Create the target output file."""
        mhcfg = self.create_mh_config()
        save_yaml(self._path, mhcfg, sort_keys=False)
        if self._path:
            logger.info(f"Created: {self._path}")
        return mhcfg



# Copyright 2021 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Virt transformer module."""

import secrets

from mrack.providers.provider import STRATEGY_ABORT
from mrack.transformers.transformer import DEFAULT_ATTEMPTS, Transformer

CONFIG_KEY = "virt"


class VirtTransformer(Transformer):
    """Virt transformer."""

    _config_key = CONFIG_KEY
    _required_config_attrs = ["images", "options", "groups"]

    def __init__(self):
        """Initialize Virt transformer."""
        self.run_id = secrets.token_urlsafe()[:6]

    async def init_provider(self):
        """Initialize associate provider and transformer display name."""
        self.dsp_name = "Virt"
        await self._provider.init(
            strategy=self.config.get("strategy", STRATEGY_ABORT),
            max_retry=self.config.get("max_retry", DEFAULT_ATTEMPTS),
        )

    def _get_host_option(self, host, name):
        default_options = self.config["options"]
        group_options = self.config["groups"].get(host["group"], {})
        val = host.get(name) or group_options.get(name) or default_options.get(name)
        if val:
            val = str(val)
        return val

    def create_host_requirement(self, host):
        """Create single input for podman provisioner."""
        req = {
            "name": host["name"],
            "os": host["os"],
            "group": host["group"],
            "run_id": self.run_id,
            "image_url": self._get_image(host),
            "ssh_path": self._get_host_option(host, "ssh_path"),
        }

        for option in [
            "ram",
            "vcpus",
            "disksize",
            "timeout",
            "vnc",
            "no_graphics",
            "keep",
        ]:
            req[option] = self._get_host_option(host, option)
        return req
